/* global.c - Global notice service
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: global.c,v 1.24.2.4 2001/06/16 16:39:32 zoot Exp $
 */

#include "config.h"
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "common.h"
#include "dict.h"
#include "conf.h"
#include "hash.h"
#include "parse.h"
#include "log.h"
#include "nickserv.h"
#include "global.h"
#include "policer.h"
#include "recdb.h"
#include "timeq.h"
#include "alias.h"

#include "helpfile.h"

#define GLOBAL_CONF_NAME	"services/global"

#define GLOBAL_DB		"global.db"
#define GLOBAL_TEMP_DB		"global.db.new"
#define GLOBAL_HELPFILE		"global.help"

/* Global options */
#define KEY_DB_BACKUP_FREQ	"db_backup_freq"
#define KEY_NICK		"nick"

/* Message data */
#define KEY_FLAGS		"flags"
#define KEY_POSTED		"posted"
#define KEY_DURATION		"duration"
#define KEY_FROM		"from"
#define KEY_MESSAGE		"message"

/* Administrative messages */
#define GMSG_WRITE_SUCCESS	"Wrote $G database in "FMT_TIME_T".%03ld seconds."
#define GMSG_HELP_READ_SUCCESS	"Read help database in "FMT_TIME_T".%03ld seconds."

#define GMSG_INVALID_ACTION	"$b%s$b is an invalid action."
#define GMSG_INVALID_TARGET	"$b%s$b is an invalid message target."
#define GMSG_MESSAGE_REQUIRED	"You $bmust$b provide a message to send."

#define GMSG_MESSAGE_SENT	"Message to $b%s$b sent."
#define GMSG_MESSAGE_ADDED	"Message to $b%s$b with ID %ld added."
#define GMSG_MESSAGE_DELETED	"Message $b%s$b deleted."
#define GMSG_ID_INVALID		"$b%s$b is an invalid message ID."

#define GMSG_MESSAGE_COUNT	"$b%d$b messages found."
#define GMSG_NO_MESSAGES	"There are no messages for you."

#define GMSG_MESSAGE_HEADER	"ID:\tTarget:\t\tExpires:\tFrom:\tMessage:"
#define GMSG_MESSAGE_DATA	"%ld\t%s\t%s\t%s\t%s"

/* Clarification: Notices are immediate, they are sent to matching users
   _once_, then forgotten. Messages are stored in Global's database and
   continually sent to users as they match the target specification until
   they are deleted. */
#define GMSG_NOTICE_SOURCE	"[$b%s$b] Notice from %s:"
#define GMSG_NOTICE_FORMAT	recipients, from

#define GMSG_MESSAGE_SOURCE	"[$b%s$b] Notice from %s, posted %s:"
#define GMSG_MESSAGE_FORMAT	recipients, from, posted

#define MESSAGELEN		20

#define GLOBAL_SYNTAX() 	global_help(user, argv[0], 0)
#define GLOBAL_FUNC(NAME) int NAME(struct userNode *user, unsigned int argc, unsigned char *argv[])
typedef GLOBAL_FUNC(global_func_t);

static GLOBAL_FUNC(cmd_help);

struct userNode *global;
struct helpfile *global_helpfile;
static dict_t global_func_dict;

static struct globalMessage *messageList = NULL;
static long messageCount = 0;

static struct
{
    unsigned long db_backup_frequency;
} global_conf;

struct global_cmd
{
    global_func_t *func;
    double weight;
};

const char *GMSG_VERSION_ID = "$Id: global.c,v 1.24.2.4 2001/06/16 16:39:32 zoot Exp $";

#define global_notice(target, format...) send_message(target , global , ## format)

static void
global_help(struct userNode *user, const char *topic, int show_aliases)
{
    char *aliases, *resolved = resolve_alias("Global", (char *)topic);

    send_help(user, global, global_helpfile, resolved);
    if(show_aliases && (aliases = command_aliases("Global", resolved)))
    {
	char *c = aliases;

	for(c = aliases; *c; c++)
	{
	    *c = toupper(*c);
	}

	global_notice(user, "Aliases: %s", aliases);
	free(aliases);
    }
}

static void
global_privmsg(struct userNode* user, char *text)
{
    struct global_cmd *cmd;
    unsigned char *argv[MAXNUMPARAMS];
    int argc;

    argc = split_line(text, false, ArrayLength(argv), argv);

    if(!argc)
    {
	return;
    }

    cmd = dict_find(global_func_dict, argv[0], NULL);

    if(!cmd)
    {
	global_notice(user, MSG_COMMAND_UNKNOWN, argv[0]);
	return;
    }

    if(!policer_conforms(user->command_policer, now, cmd->weight))
    {
	global_notice(user, MSG_COMMAND_FLOOD);
	return;
    }

    cmd->func(user, argc, argv);
}

void message_expire(void *data);

static struct globalMessage*
message_add(long flags, time_t posted, unsigned long duration, char *from, char *msg)
{
    struct globalMessage *message;

    message = malloc(sizeof(struct globalMessage));

    if(!message)
    {
	return NULL;
    }

    message->id = messageCount++;
    message->flags = flags;
    message->posted = posted;
    message->duration = duration;
    message->from = strdup(from);
    message->message = strdup(msg);

    if(messageList)
    {
	messageList->prev = message;
    }

    message->prev = NULL;
    message->next = messageList;

    messageList = message;

    if(duration)
    {
	timeq_add(now + duration, message_expire, message);
    }

    return message;
}

static void
message_del(struct globalMessage *message)
{
    if(message->duration)
    {
	timeq_del(0, NULL, message, TIMEQ_IGNORE_FUNC | TIMEQ_IGNORE_WHEN);
    }

    if(message->prev) message->prev->next = message->next;
    else messageList = message->next;

    if(message->next) message->next->prev = message->prev;

    if(message->from) free(message->from);
    if(message->message) free(message->message);

    free(message);
}

void message_expire(void *data)
{
    struct globalMessage *message = data;

    message->duration = 0;
    message_del(message);
}

static struct globalMessage*
message_create(struct userNode *user, unsigned int argc, unsigned char *argv[])
{
    unsigned long duration = 0;
    char *text = NULL;
    long flags = -1;
    unsigned int i;

    for(i = 0; i < argc; i++)
    {
	if((i + 1) > argc)
	{
	    global_notice(user, MSG_MISSING_PARAMS, argv[argc]);
	    return NULL;
	}

	if(!strcasecmp(argv[i], "text"))
	{
	    i++;
	    unsplit_string(argv + i, argc - i);
	    text = argv[i];
	    break;
	} else if (!strcasecmp(argv[i], "sourceless")) {
	    i++;
	    flags |= MESSAGE_OPTION_SOURCELESS;
	} else if (!strcasecmp(argv[i], "target")) {
	    i++;

	    if(!strcasecmp(argv[i], "all")) {
		flags |= MESSAGE_RECIPIENT_ALL;
	    } else if(!strcasecmp(argv[i], "users")) {
		flags |= MESSAGE_RECIPIENT_LUSERS;
	    } else if(!strcasecmp(argv[i], "helpers")) {
		flags |= MESSAGE_RECIPIENT_HELPERS;
	    } else if(!strcasecmp(argv[i], "opers")) {
		flags |= MESSAGE_RECIPIENT_OPERS;
	    } else if(!strcasecmp(argv[i], "staff")) {
		flags |= MESSAGE_RECIPIENT_STAFF;
	    } else if(!strcasecmp(argv[i], "channels")) {
		flags |= MESSAGE_RECIPIENT_CHANNELS;
	    } else {
		global_notice(user, GMSG_INVALID_TARGET, argv[i]);
		return NULL;
	    }
	} else if (strcasecmp(argv[i], "duration") == 0) {
	    duration = ParseInterval(argv[++i]);
	} else {
	    global_notice(user, MSG_INVALID_CRITERIA, argv[i]);
	    return NULL;
	}
    }

    if(flags <= 0)
    {
	flags = MESSAGE_RECIPIENT_LUSERS;
    }

    if(!text) {
	global_notice(user, GMSG_MESSAGE_REQUIRED);
	return NULL;
    }

    return message_add(flags, now, duration, user->handle_info->handle, text);
}

static const char *
messageType(const struct globalMessage *message)
{
    if((message->flags & MESSAGE_RECIPIENT_ALL) == MESSAGE_RECIPIENT_ALL)
    {
	return "all";
    }
    if((message->flags & MESSAGE_RECIPIENT_STAFF) == MESSAGE_RECIPIENT_STAFF)
    {
	return "staff";
    }
    else if(message->flags & MESSAGE_RECIPIENT_OPERS)
    {
	return "opers";
    }
    else if(message->flags & MESSAGE_RECIPIENT_HELPERS)
    {
	return "helpers";
    }
    else if(message->flags & MESSAGE_RECIPIENT_LUSERS)
    {
	return "users";
    }
    else
    {
	return "channels";
    }
}

static void
notice_target(const char *target, struct globalMessage *message)
{
    const char *recipients = messageType(message), *from = message->from;

    if(!(message->flags & MESSAGE_OPTION_SOURCELESS))
    {
	if(message->flags & MESSAGE_OPTION_IMMEDIATE)
	{
	    send_target_message(0, target, global, GMSG_NOTICE_SOURCE, GMSG_NOTICE_FORMAT);
	}
	else
	{
	    char posted[24];
	    struct tm tm;

	    localtime_r(&message->posted, &tm);
	    strftime(posted, sizeof(posted), "%I:%M %p, %m/%d/%Y", &tm);

	    send_target_message(0, target, global, GMSG_MESSAGE_SOURCE, GMSG_MESSAGE_FORMAT);
	}
    }

    send_target_message(0, target, global, "%s", message->message);
}

#define notice_user(USER, MESSAGE) notice_target((USER)->numeric, MESSAGE)

static int
notice_channel(const char *key, void *data, void *extra)
{
    struct chanNode *channel = data;

    /* It should be safe to assume channel is not NULL. */
    if(channel->channel_info)
    {
	notice_target(key, extra);
    }

    return 0;
}

static void
message_send(struct globalMessage *message)
{
    struct userNode *user;
    unsigned long n;

    if(message->flags & MESSAGE_RECIPIENT_CHANNELS)
    {
	dict_foreach(channels, notice_channel, message);
    }

    if(message->flags & MESSAGE_RECIPIENT_LUSERS)
    {
	notice_target("$*", message);
	return;
    }

    if(message->flags & MESSAGE_RECIPIENT_OPERS)
    {
	for(n = 0; n < curr_opers.used; n++)
	{
	    user = curr_opers.list[n];

	    if(user->uplink != self)
	    {
		notice_user(user, message);
	    }
	}
    }

    if(message->flags & MESSAGE_RECIPIENT_HELPERS)
    {
	for(n = 0; n < curr_helpers.used; n++)
	{
	    user = curr_helpers.list[n];
	    notice_user(user, message);
	}
    }
}

void
global_message(long targets, char *text)
{
    struct globalMessage *message;

    if(!targets)
    {
	return;
    }

    message = message_add(targets | MESSAGE_OPTION_SOURCELESS, now, 0, "", text);

    if(!message)
    {
	return;
    }

    message_send(message);
    message_del(message);
}

static GLOBAL_FUNC(cmd_notice)
{
    struct globalMessage *message = NULL;
    const char *recipient = NULL;
    long target = 0;

    if(!IsOper(user))
    {
	global_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    if(!user->handle_info)
    {
	if(!get_handle_info(user->nick) && !get_nick_info(user->nick))
	{
	    global_notice(user, MSG_REGISTER_HANDLE);
	    return 0;
	}

	global_notice(user, MSG_AUTHENTICATE);
	return 0;
    }

    if(argc < 3)
    {
	global_notice(user, MSG_MISSING_PARAMS, argv[0]);
	GLOBAL_SYNTAX();
	return 0;
    }

    if(!strcmp(argv[1], "all")) {
	target = MESSAGE_RECIPIENT_ALL;
    } else if(!strcmp(argv[1], "users")) {
	target = MESSAGE_RECIPIENT_LUSERS;
    } else if(!strcmp(argv[1], "helpers")) {
	target = MESSAGE_RECIPIENT_HELPERS;
    } else if(!strcmp(argv[1], "opers")) {
	target = MESSAGE_RECIPIENT_OPERS;
    } else if(!strcmp(argv[1], "privileged")) {
	target |= MESSAGE_RECIPIENT_HELPERS | MESSAGE_RECIPIENT_OPERS;
    } else if(!strcmp(argv[1], "channels")) {
	target = MESSAGE_RECIPIENT_CHANNELS;
    } else {
	global_notice(user, GMSG_INVALID_TARGET, argv[1]);
	return 0;
    }

    unsplit_string(argv + 2, argc - 2);
    message = message_add(target | MESSAGE_OPTION_IMMEDIATE, now, 0, user->handle_info->handle, argv[2]);

    if(!message)
    {
	return 0;
    }

    recipient = messageType(message);

    message_send(message);
    message_del(message);

    log(G_LOG, LOG_INFO, "%s sent a notice to %s.\n", user->handle_info->handle, recipient);
    global_notice(user, GMSG_MESSAGE_SENT, recipient);
    return 1;
}

static GLOBAL_FUNC(cmd_message)
{
    struct globalMessage *message = NULL;
    const char *recipient = NULL;

    if(!IsOper(user))
    {
	global_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    if(!user->handle_info)
    {
	if(!get_handle_info(user->nick) && !get_nick_info(user->nick))
	{
	    global_notice(user, MSG_REGISTER_HANDLE);
	    return 0;
	}

	global_notice(user, MSG_AUTHENTICATE);
	return 0;
    }

    if(argc < 3)
    {
	global_notice(user, MSG_MISSING_PARAMS, argv[0]);
	GLOBAL_SYNTAX();
	return 0;
    }

    message = message_create(user, argc - 1, argv + 1);
    if(!message)
    {
	return 0;
    }

    recipient = messageType(message);

    log(G_LOG, LOG_INFO, "%s sent a notice to %s.\n", user->handle_info->handle, recipient);
    global_notice(user, GMSG_MESSAGE_ADDED, recipient, message->id);
    return 1;
}

static GLOBAL_FUNC(cmd_list)
{
    struct globalMessage *message = NULL;
    const char *recipient = NULL;
    char expires[INTERVALLEN], msg[MESSAGELEN];
    unsigned int length;
    (void)user;(void)argv;(void)argc;

    if(!IsOper(user))
    {
	global_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    global_notice(user, GMSG_MESSAGE_HEADER);

    if(!messageList)
    {
	global_notice(user, "$bNone.$b");
    }

    for(message = messageList; message; message = message->next)
    {
	recipient = messageType(message);
	intervalString(expires, message->posted + message->duration - now);

	length = strlen(message->message);
	safestrncpy(msg, message->message, sizeof(msg));

	if(length >= MESSAGELEN)
	{
	    msg[MESSAGELEN - 1] = 0;
	    msg[MESSAGELEN - 2] = msg[MESSAGELEN - 3] = msg[MESSAGELEN - 4] = '.';
	}

        global_notice(user, GMSG_MESSAGE_DATA, message->id, recipient, message->duration ? expires : "Never.\t", message->from, msg);
    }

    return 1;
}

static GLOBAL_FUNC(cmd_remove)
{
    struct globalMessage *message = NULL;
    unsigned long id;

    if(!IsOper(user))
    {
	global_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    if(!user->handle_info)
    {
	if(!get_handle_info(user->nick) && !get_nick_info(user->nick))
	{
	    global_notice(user, MSG_REGISTER_HANDLE);
	    return 0;
	}

	global_notice(user, MSG_AUTHENTICATE);
	return 0;
    }

    if(argc < 2)
    {
	global_notice(user, MSG_MISSING_PARAMS, argv[0]);
	GLOBAL_SYNTAX();
	return 0;
    }

    id = strtoul(argv[1], NULL, 0);

    for(message = messageList; message; message = message->next)
    {
	if(message->id == id)
	{
	    message_del(message);
	    global_notice(user, GMSG_MESSAGE_DELETED, argv[1]);
	    log(G_LOG, LOG_INFO, "%s removed a notice with ID %li.\n", user->handle_info->handle, id);
	    return 1;
	}
    }

    global_notice(user, GMSG_ID_INVALID, argv[1]);
    return 1;
}

static unsigned int
send_messages(struct userNode *user, long mask)
{
    struct globalMessage *message = messageList;
    unsigned int count = 0;

    while(message)
    {
	if(message->flags & mask)
	{
	    notice_user(user, message);
	    count++;
	}

	message = message->next;
    }

    return count;
}

static GLOBAL_FUNC(cmd_messages)
{
    long mask = MESSAGE_RECIPIENT_LUSERS | MESSAGE_RECIPIENT_CHANNELS;
    unsigned int count;
    (void)argc;(void)argv;

    if(IsOper(user))
    {
	mask |= MESSAGE_RECIPIENT_OPERS;
    }

    if(user->handle_info && HANDLE_FLAGGED(user->handle_info, HELPER))
    {
	mask |= MESSAGE_RECIPIENT_HELPERS;
    }

    count = send_messages(user, mask);
    if(count)
    {
	global_notice(user, GMSG_MESSAGE_COUNT, count);
    }
    else
    {
	global_notice(user, GMSG_NO_MESSAGES);
    }

    return 1;
}

static GLOBAL_FUNC(cmd_version)
{
    (void)argc;(void)argv;

    global_notice(user, MSG_VERSION);
    global_notice(user, GMSG_VERSION_ID);

    return 1;
}

static int
global_process_user(struct userNode *user)
{
    /* TODO: policers for global motd/memos may be useful. */
    send_messages(user, MESSAGE_RECIPIENT_LUSERS);
    return 0;
}

static void
global_process_auth(struct userNode *user, struct handle_info *old_handle)
{
    (void)old_handle;
    if(user->handle_info && HANDLE_FLAGGED(user->handle_info, HELPER))
    {
	send_messages(user, MESSAGE_RECIPIENT_HELPERS);
    }
}

static void
global_process_oper(struct userNode *user)
{
    send_messages(user, MESSAGE_RECIPIENT_OPERS);
}

static void
global_helpfile_read(void)
{
    global_helpfile = open_helpfile(GLOBAL_HELPFILE, NULL);
}

static GLOBAL_FUNC(cmd_help)
{
    const char *topic;

    if (argc < 2) {
	topic = NULL;
    } else {
	topic = argv[1];
	unsplit_string(argv+1, argc-1);
    }
    global_help(user, topic, 1);
    return 1;
}

static GLOBAL_FUNC(cmd_readhelp)
{
    struct timeval start, stop;
    struct helpfile *old_helpfile = global_helpfile;
    (void)argc; (void)argv;

    if(!IsOper(user))
    {
	global_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    gettimeofday(&start, NULL);
    global_helpfile_read();
    if (global_helpfile) {
	close_helpfile(old_helpfile);
    } else {
	global_helpfile = old_helpfile;
    }
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    global_notice(user, GMSG_HELP_READ_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    log(G_LOG, LOG_INFO, "%s ordered a help file read.\n", user->handle_info->handle);

    return 1;
}

static void
global_conf_read(void)
{
    struct record_data *rd;
    dict_t conf_node;
    const char *str;

    rd = conf_get_node(GLOBAL_CONF_NAME);

    if (!rd || rd->type != RECDB_OBJECT) {
	log(G_LOG, LOG_ERROR, "config node `%s' is missing or has wrong type\n", GLOBAL_CONF_NAME);
	return;
    }

    conf_node = rd->d.object;

    str = database_get_data(conf_node, KEY_DB_BACKUP_FREQ, RECDB_QSTRING);
    global_conf.db_backup_frequency = str ? ParseInterval(str) : 7200;

    str = database_get_data(conf_node, KEY_NICK, RECDB_QSTRING);
    if(!str) return;
    NickChange(global, str);
}

int global_command_search(const char *command)
{
    struct global_cmd *cmd = dict_find(global_func_dict, command, NULL);
    return cmd ? 1 : 0;
}

int global_add_alias(const char *alias, const char *command)
{
    struct global_cmd *cmd = dict_find(global_func_dict, command, NULL);

    if(!cmd) return 0;

    dict_insert(global_func_dict, alias, cmd);
    return 1;
}

int global_del_alias(const char *alias)
{
    return dict_remove(global_func_dict, alias);
}

static struct global_cmd *
global_define_func(const char *name, global_func_t *func)
{
    struct global_cmd *cmd = malloc(sizeof(struct global_cmd));
    cmd->weight = 1.0;
    cmd->func = func;
    dict_insert(global_func_dict, name, cmd);
    return cmd;
}

static void
global_db_read(void)
{
    struct record_data *hir;
    time_t posted = 0;
    long flags = 0;
    unsigned long duration = 0;
    char *id, *str, *from, *message;

    FILE *db = fopen(GLOBAL_DB, "r");
    if (!db) return;

    while(!feof(db) && parse_record(db, &id, &hir))
    {
	if(feof(db))
	{
	    break;
	}

	if(hir->type != RECDB_OBJECT)
	{
	    log(G_LOG, LOG_WARNING, "unexpected rectype %d for %s\n", hir->type, id);
	    goto next_record;
	}

        str = database_get_data(hir->d.object, KEY_FLAGS, RECDB_QSTRING);
        flags = str ? strtoul(str, NULL, 0) : 0;

        str = database_get_data(hir->d.object, KEY_POSTED, RECDB_QSTRING);
        posted = str ? strtoul(str, NULL, 0) : 0;

        str = database_get_data(hir->d.object, KEY_DURATION, RECDB_QSTRING);
        duration = str ? strtoul(str, NULL, 0) : 0;

        from = database_get_data(hir->d.object, KEY_FROM, RECDB_QSTRING);
        message = database_get_data(hir->d.object, KEY_MESSAGE, RECDB_QSTRING);

	message_add(flags, posted, duration, from, message);

    next_record:
	free_record_data(hir);
    }

    fclose(db);
}

static void
write_message(FILE *db, struct globalMessage *message)
{
    dict_t obj = alloc_object();
    struct record_data *rd;
    char *str;

    str = alloca(12);
    snprintf(str, 12, "%li", message->flags);
    dict_insert(obj, KEY_FLAGS, alloc_record_data_qstring(str));

    snprintf(str, 12, FMT_TIME_T, message->posted);
    dict_insert(obj, KEY_POSTED, alloc_record_data_qstring(str));
    
    snprintf(str, 12, "%li", message->duration);
    dict_insert(obj, KEY_DURATION, alloc_record_data_qstring(str));

    dict_insert(obj, KEY_FROM, alloc_record_data_qstring(message->from));
    dict_insert(obj, KEY_MESSAGE, alloc_record_data_qstring(message->message));

    /* the key is ignored when read in, so cheat and use "message." */
    rd = alloc_record_data_object(obj);
    write_record(db, "message", rd);
    free_record_data(rd);
    return;
}

static void
global_db_write(void)
{
    FILE *f = fopen(GLOBAL_TEMP_DB, "w");
    struct globalMessage *message;

    for(message = messageList; message; message = message->next)
    {
        write_message(f, message);
    }

    fclose(f);
    if (rename(GLOBAL_TEMP_DB, GLOBAL_DB) == -1) {
	log(G_LOG, LOG_ERROR, "Unable to rename new db to old: %s\n", strerror(errno));
    }
}

void
global_timed_write(void *data)
{
    (void)data;

    global_db_write();
    log(G_LOG, LOG_INFO, "Successfully made scheduled backup of Global database.\n");
    timeq_add(now + global_conf.db_backup_frequency, global_timed_write, NULL);
}

static GLOBAL_FUNC(cmd_write)
{
    struct timeval start, stop;
    (void)argc;(void)argv;

    if (!IsOper(user))
    {
	global_notice(user, MSG_COMMAND_PRIVILEGED, argv[0]);
	return 0;
    }

    gettimeofday(&start, NULL);
    global_db_write();
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    global_notice(user, GMSG_WRITE_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    log(G_LOG, LOG_INFO, "%s ordered a Global database write.\n", user->nick);
    return 1;
}

void
init_global(const char *nick)
{
    global = AddService(nick, "Global Services");
    reg_privmsg_func(global->num_local, global_privmsg);
    reg_new_user_func(global_process_user);
    reg_auth_func(global_process_auth);
    reg_oper_func(global_process_oper);

    global_conf_read();
    conf_register_reload(global_conf_read);

    global_func_dict = dict_new();
    global_define_func("HELP", cmd_help);

    global_define_func("NOTICE", cmd_notice);
    global_define_func("MESSAGE", cmd_message);
    global_define_func("LIST", cmd_list);
    global_define_func("REMOVE", cmd_remove);

    global_define_func("MESSAGES", cmd_messages);
    global_define_func("VERSION", cmd_version);

    global_define_func("READHELP", cmd_readhelp);
    global_define_func("WRITE", cmd_write);

    alias_register_service("Global", global_command_search, global_add_alias, global_del_alias);

    global_db_read();
    global_helpfile_read();

    if(global_conf.db_backup_frequency) timeq_add(now + global_conf.db_backup_frequency, global_timed_write, NULL);

    reg_db_write_func(global_db_write);
}
